<?php
/********************************* PHP-P10 ******************************
 *    P10 uplink class by pk910   (c)2011 pk910                         *
 ************************************************************************
 *                          Version 2 (OOP)                             *
 *                                                                      *
 * PHP-P10 is free software; you can redistribute it and/or modify      *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with PHP-P10; if not, write to the Free Software Foundation,   *
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.       *
 *                                                                      *
 ************************************************************************
 * 
 *  Uplink/Uplink.class.php
 *
 * This file contains the basic P10 Protocol handler.
 *
 ************************************************************************
 * accessable methods:
 *
 * void initialize()
 *     has to be called after the settings have been set.
 *
 * void loop() 
 *     loop function that should be calles as many times as possible. 
 *     It reads from the socket and BLOCKS the script execution for a 
 *     specific time if nothing is received.
 *
 * void setUplinkHost(String $host, int $port, bool $ssl = false, String $bind = null)
 *     sets the Uplink connection information.
 *
 * void setLoopTimeout(int $timeout)
 *     sets the maximum time loop() is blocking the script execution.
 *
 * void setUplinkServer(int $numeric, String $name, String $password, String $description)
 *     sets the own P10 Server information.
 *
 * void setValidateServer(String $name, String $password)
 *     sets additional security relevant information about the remote server.
 *
 * void setHISOptions(String $serverName, String $serverDescription String $usermask)
 *     sets the default Account fakehost
 *
 * void setEventHandler(EventHandler $event_handler)
 *     sets the EventHandlder
 */
require_once("Client.class.php");
require_once("Numerics.class.php");
require_once("P10Formatter.class.php");
require_once("P10_Server.class.php");
require_once("P10_User.class.php");
require_once("P10_Channel.class.php");
require_once("P10_ModeSets.class.php");
require_once("EventHandler.interface.php");
require_once("IPAddr.class.php");

$e=1;
define("ERR_NICK_IN_USE", $e++);
define("ERR_INVALID_USER", $e++);
define("ERR_INVALID_CHANNAME", $e++);
define("ERR_NOT_CONNECTED", $e++);
define("ERR_NOT_ON_CHANNEL", $e++);

class Uplink {
	private $client;
	private $settings = array();
	private $server;
	private $eventHandler = null;
	private $last_local_numeric = 1;
	
	const FLAG_P10SESSION      = 0x0001; //connection is in P10 mode (server is connected)
	const FLAG_SECURITY_QUIT   = 0x0002; //local connection abort because of security issues
	const FLAG_NOT_CONNECTABLE = 0x0004; //remote server is not connectable
	const FLAG_BURST_PENDING   = 0x0008; //we still have to burst
	const FLAG_CONNECTED       = 0x0010; //connected and synced (ready)
	const FLAG_GOT_PASS        = 0x0020; //got PASS from the remote Server
	private $flags = 0;
	
	public function __construct() {
		$this->client = new Client();
		$this->setSetting("recv_timeout", 1000);
		$this->setSetting("his_usermask", "user.NoMask");
	}
	
	public function initialize() {
		if($this->server) {
			trigger_error("Uplink already initialized.", E_USER_ERROR);
			return;
		}
		$self_numeric = $this->getSetting("numeric");
		$self_name = $this->getSetting("name");
		$self_description = $this->getSetting("description");
		if(!$self_numeric || !$self_name) {
			trigger_error("Server Settings missing.", E_USER_ERROR);
			return;
		}
		$this->server = new P10_Server($self_name, $self_numeric, null, time(), time(), $self_description);
	}
	
	public function loop() {
		if($this->server == null) {
			trigger_error("Uplink not initialized.", E_USER_ERROR);
			return;
		}
		if(!$this->client->connected()) {
			if(($this->flags & self::FLAG_P10SESSION)) {
				//Server got disconnected
				$this->server->disconnectServer($this->eventHandler, true);
				$this->flags &= ~self::FLAG_P10SESSION;
			}
			$host = $this->getSetting("host");
			$port = $this->getSetting("port");
			if($host == null || $port == null) {
				trigger_error("Uplink Settings missing.", E_USER_ERROR);
				return;
			}
			if(($this->flags & self::FLAG_SECURITY_QUIT) || ($this->flags & self::FLAG_NOT_CONNECTABLE)) {
				sleep(1);
			}
			$state = $this->client->connect($host, $port, $this->getSetting("bind"), $this->getSetting("ssl"), $this->getSetting("recv_timeout"));
			if(!$state) {
				usleep($this->getSetting("recv_timeout") / 1000);
				$this->flags |= self::FLAG_NOT_CONNECTABLE;
				return;
			}
			$this->flags = 0;
			$this->loginServer();
		}
		//try to receive new data from the Uplink
		$lines = $this->client->recv();
		if($lines == null) return;
		foreach($lines as $line) {
			$this->parseLine($line);
		}
	}
	
	public function shutdown() {
		if($this->client->connected()) {
			if(($this->flags & self::FLAG_P10SESSION)) {
				$this->send("SQ", "Shutdown requested.");
			}
			$this->client->disconnect();
		}
	}
	
	public function setUplinkHost($host, $port, $ssl = false, $bind = null) {
		$this->setSetting("host", $host);
		$this->setSetting("port", $port);
		$this->setSetting("ssl", $ssl);
		$this->setSetting("bind", $bind);
	}
	
	public function setLoopTimeout($timeout) {
		$this->setSetting("recv_timeout", $timeout);
	}
	
	public function setUplinkServer($numeric, $name, $password, $description) {
		$this->setSetting("numeric", Numerics::intToNum($numeric, 2));
		$this->setSetting("name", $name);
		$this->setSetting("password", $password);
		$this->setSetting("description", $description);
	}
	
	public function setValidateServer($name, $password) {
		$this->setSetting("their_name", $name);
		$this->setSetting("their_password", $password);
	}
	
	public function setHISOptions($servername, $serverdesc, $usermask) {
		$this->setSetting("his_name", $servername);
		$this->setSetting("his_desc", $serverdesc);
		$this->setSetting("his_usermask", $usermask);
	}
	
	public function setEventHandler($event_handler) {
		if(!is_a($event_handler, "EventHandler")) {
			trigger_error((is_object($event_handler) ? get_class($event_handler) : gettype($event_handler))." is NOT a valid EventHandler.", E_USER_ERROR);
			return;
		}
		$this->eventHandler = $event_handler;
	}
	
	private function setSetting($setting, $value) {
		$this->settings[$setting] = $value;
	}
	
	private function getSetting($setting) {
		if(array_key_exists($setting, $this->settings)) {
			return $this->settings[$setting];
		} else {
			return null;
		}
	}
	
	private function loginServer() {
		$password = $this->getSetting("password");
		$this->send("PASS", $password);
		$this->send("SERVER", $this->server->getName(), $this->server->getStartTime(), $this->server->getLinkTime(), $this->server->getNumeric(), $this->server->getDescription());
	}
	
	private function parseLine($line) {
		$highExplode = explode(" :", $line, 2);
		$tokens = explode(" ", $highExplode[0]);
		if(count($highExplode) > 1)
			$tokens[] = $highExplode[1];
		$cmdPos = (($this->flags & self::FLAG_P10SESSION) ? 1 : 0);
		if($cmdPos == 1) $from = $tokens[0];
		else $from = null;
		$arguments = array_slice($tokens, $cmdPos + 1);
		if(($this->flags & self::FLAG_P10SESSION) && $this->eventHandler) {
			$this->eventHandler->event_preparse($from, strtoupper($tokens[$cmdPos]), $arguments);
		}
		switch(strtoupper($tokens[$cmdPos])) {
		//pre P10 Session
			case "PASS":
				$this->recv_pass($from, $arguments);
				break;
			case "SERVER":
				$this->recv_server($from, $arguments);
				break;
			case "ERROR":
				$this->recv_error($from, $arguments);
				break;
		//P10 Session
			case "S":
				$this->recv_server($from, $arguments);
				break;
			case "G":
				$this->recv_ping($from, $arguments);
				break;
			case "N":
				$this->recv_nick($from, $arguments);
				break;
			case "EB":
				$this->recv_end_of_burst($from, $arguments);
				break;
			case "EA":
				$this->recv_end_of_burst_ack($from, $arguments);
				break;
			case "SQ":
				$this->recv_server_quit($from, $arguments);
				break;
			case "Q":
				$this->recv_quit($from, $arguments);
				break;
			case "B":
				$this->recv_burst($from, $arguments);
				break;
			case "J":
			case "C":
				$this->recv_join($from, $arguments);
				break;
			case "L":
				$this->recv_part($from, $arguments);
				break;
			case "K":
				$this->recv_kick($from, $arguments);
				break;
			case "D":
				$this->recv_kill($from, $arguments);
				break;
			case "P":
				$this->recv_privmsg($from, $arguments);
				break;
			case "O":
				$this->recv_notice($from, $arguments);
				break;
			case "W":
				$this->recv_whois($from, $arguments);
				break;
			case "A":
				$this->recv_away($from, $arguments);
				break;
			case "M":
			case "OM":
				$this->recv_mode($from, $arguments);
				break;
		//default
			default:
				//unknown cmd
				if($this->eventHandler)
					$this->eventHandler->event_unknown_cmd($from, strtoupper($tokens[$cmdPos]), $arguments);
				break;
		}
	}
	
	private function send($command) {
		if(func_num_args() > 1) {
			$args = array_slice(func_get_args(), 1);
			$command = P10Formatter::formatCMD($this->getSetting("numeric"), $command, $args);
		} else {
			$command = P10Formatter::formatCMD($this->getSetting("numeric"), $command, array());
		}
		$this->client->send($command);
	}
	
	/********************************************************************************************
	 *                                     INCOMING COMMANDS                                    *
	 ********************************************************************************************/
	
	private function recv_pass($from, $args) {
		$their_pass = $this->getSetting("their_password");
		if($their_pass) {
			if($args[0] != $their_pass) {
				//security QUIT
				$this->flags |= self::FLAG_SECURITY_QUIT;
				$this->send("ERROR", "Incorrect password received.");
				$this->client->disconnect();
				return;
			}
			$this->flags |= self::FLAG_GOT_PASS;
		}
	}
	
	private function recv_error($from, $args) {
		//we received an error - the socket is dead for us, now
		//maybe we want to log this, later
		$this->client->disconnect();
	}
	
	private function recv_server($from, $args) {
		if($from == null) {
			//our uplink Server
			$their_name = $this->getSetting("their_name");
			if($their_name && $args[0] != $their_name) {
				$this->flags |= self::FLAG_SECURITY_QUIT;
				$this->send("ERROR", "Invalid Server name");
				$this->client->disconnect();
				return;
			}
			if($this->getSetting("their_password") && !($this->flags & self::FLAG_GOT_PASS)) {
				$this->flags |= self::FLAG_SECURITY_QUIT;
				$this->send("ERROR", "PASS missing.");
				$this->client->disconnect();
				return;
			}
			$new_server = new P10_Server($args[0], substr($args[5],0,2), $this->server, $args[2], $args[3], $args[7]);
			$this->server->addServer($new_server);
			$this->flags |= self::FLAG_P10SESSION | self::FLAG_BURST_PENDING;
			if($this->eventHandler)
				$this->eventHandler->event_newserver($new_server, !($this->flags & self::FLAG_CONNECTED));
		} else {
			//another server got a new slave server ^^
			$server = P10_Server::getServerByNum($from);
			if($server == null) {
				trigger_error("Parent Server (".$from.") does not exist or was not found on recv_server.", E_USER_ERROR);
				return;
			}
			$new_server = new P10_Server($args[0], substr($args[5],0,2), $server, $args[2], $args[3], $args[7]);
			$server->addServer($new_server);
			if($this->eventHandler)
				$this->eventHandler->event_newserver($new_server, !($this->flags & self::FLAG_CONNECTED));
		}
	}
	
	private function recv_ping($from, $args) {
		$this->send("Z", $args[0]); //simply PONG
		P10_Channel::recheckAllChannels();
	}
	
	private function recv_nick($from, $args) {
		if(count($args) <= 2) {
			//Nick change
			$user = P10_User::getUserByNum($from);
			if($user == null) {
				trigger_error("Server tries to change the nick of an user that does not exist or was not found on recv_nick.", E_USER_ERROR);
				return;
			}
			if($this->eventHandler)
				$this->eventHandler->event_nick($user, $args[0]);
			$user->setNick($args[0]);
		} else {
			//New User
			$numeric = $args[count($args)-2];
			$nick = $args[0];
			$server = P10_Server::getServerByNum($from);
			if($server == null) {
				trigger_error("Server (".$from.") the User (".$nick.") is coming from does not exist or was not found on recv_nick.", E_USER_ERROR);
				return;
			}
			if(substr($numeric,0,2) != $from) {
				trigger_error("A Server (".$from.") tries to connect a User with an invalid User numeric ('".$numeric."' does not belong to the Server)", E_USER_WARNING);
			}
			$connect_time = $args[2];
			$ident = $args[3];
			$host = $args[4];
			$modes = implode(" ",array_slice($args, 5, count($args)-8));
			$modes = new P10_UserModeSet($modes);
			$ip = new IPAddr($args[count($args)-3]);
			$realname = $args[count($args)-1];
			$user = new P10_User($nick, $numeric, $server, $connect_time, $ident, $host, $ip, $realname, $modes);
			if($this->eventHandler)
				$this->eventHandler->event_connect($user, !($this->flags & self::FLAG_CONNECTED));
		}
	}
	
	private function recv_end_of_burst($from, $args) {
		if(($this->flags & self::FLAG_BURST_PENDING)) {
			$this->burst();
			$this->send("EA");
			$this->flags &= ~self::FLAG_BURST_PENDING;
		}
	}
	
	private function recv_end_of_burst_ack($from, $args) {
		$this->flags |= self::FLAG_CONNECTED;
	}
	
	private function recv_server_quit($from, $args) {
		$server = P10_Server::getServerByName($args[0]);
		if($server == null) {
			trigger_error("Server (".$args[0].") not found.", E_USER_ERROR);
			return;
		}
		$server->disconnectServer($this->eventHandler);
	}
	
	private function recv_quit($from, $args) {
		$user = P10_User::getUserByNum($from);
		if($user == null) {
			trigger_error("Server tries to quit an user that does not exist or was not found on recv_quit.", E_USER_ERROR);
			return;
		}
		if($this->eventHandler)
			$this->eventHandler->event_quit($user, $args[0]);
		$user->quit($args[0]);
	}
	
	private function recv_burst($from, $args) {
		$name = $args[0];
		$create_time = $args[1];
		$channel = P10_Channel::getChannelByName($name);
		if($channel == null)
			$channel = new P10_Channel($name);
		$channel->setCreateTime($create_time);
		$modes = $channel->getModes();
		$userstr = $args[count($args)-1];
		$modeparamcount = count($args)-3;
		if($userstr[0] == "%") {
			//ban list
			$banlist = explode(" ", substr($userstr, 1));
			foreach($banlist as $ban) {
				//TODO: save bans
			}
			$userstr = $args[count($args)-2];
			$modeparamcount--;
		}
		if($userstr[0] == "+") { //MODE String
			$modeparamcount++;
			$userstr = "";
		}
		$users = explode(",",$userstr);
		$isop = false; $isvoice = false;
		foreach($users as $user) {
			if($user == "") continue;
			$uexp = explode(":", $user);
			if(strlen($uexp[0]) != 5) {
				trigger_error("burst parse error: '".$uexp[0]."' is not an user numeric.", E_USER_ERROR);
				return;
			}
			if(count($uexp) > 1) {
				$isop = false;
				$isvoice = false;
				for($i = 0; $i < strlen($uexp[1]); $i++) {
					if($uexp[1][0] == "@") $isop = true;
					if($uexp[1][0] == "+") $isvoice = true;
				}
			}
			$user = P10_User::getUserByNum($uexp[0]);
			if($user == null) {
				trigger_error("burst parse error: cant find User '".$uexp[0]."'.", E_USER_ERROR);
				return;
			}
			$channel->burstUser($user, $isop, $isvoice);
			if($this->eventHandler)
				$this->eventHandler->event_join($user, $channel, true);
		}
		$modes->parseModes(implode(" ", array_slice($args, 2, $modeparamcount)));
	}
	
	private function recv_join($from, $args) {
		$user = P10_User::getUserByNum($from);
		if($user == null) {
			trigger_error("Server tries to join an user that does not exist or was not found on recv_join.", E_USER_ERROR);
			return;
		}
		$channel = P10_Channel::getChannelByName($args[0]);
		if($channel == null)
			$channel = new P10_Channel($args[0]);
		$channel->joinUser($user);
		if($this->eventHandler)
			$this->eventHandler->event_join($user, $channel, false);
	}
	
	private function recv_part($from, $args) {
		$user = P10_User::getUserByNum($from);
		if($user == null) {
			trigger_error("Server tries to part an user that does not exist or was not found on recv_join.", E_USER_ERROR);
			return;
		}
		$channel = P10_Channel::getChannelByName($args[0]);
		if($channel == null)
			$channel = new P10_Channel($args[0]);
		if($this->eventHandler)
			$this->eventHandler->event_part($user, $channel, $args[1]);
		$channel->partUser($user);
	}
	
	private function recv_kick($from, $args) {
		$user = P10_User::getUserByNum($from);
		if($user == null) {
			trigger_error("An unknown user tries to kick another user on recv_kick.", E_USER_ERROR);
			return;
		}
		$channel = P10_Channel::getChannelByName($args[0]);
		if($channel == null)
			$channel = new P10_Channel($args[0]);
		$target = P10_User::getUserByNum($args[1]);
		if($target == null) {
			trigger_error("Someone tries to kick an user that does not exist or was not found on recv_kick.", E_USER_ERROR);
			return;
		}
		if($this->eventHandler)
			$this->eventHandler->event_kick($user, $target, $channel, $args[1]);
		$channel->partUser($user);
	}
	
	private function recv_kill($from, $args) {
		$user = P10_User::getUserByNum($args[0]);
		if($user == null) {
			trigger_error("Server tries to kill an user that does not exist or was not found on recv_quit.", E_USER_ERROR);
			return;
		}
		if($this->eventHandler)
			$this->eventHandler->event_quit($user, "Killed (".$args[1].")");
		$user->quit($args[1]);
	}
	
	private function recv_privmsg($from, $args) {
		$user = P10_User::getUserByNum($from);
		if($user == null) {
			trigger_error("Server tries to send a privmsg from an user that does not exist or was not found on recv_privmsg.", E_USER_ERROR);
			return;
		}
		if($this->eventHandler) {
			if($args[0][0] == "#") {
				$channel = P10_Channel::getChannelByName($args[0]);
				if($channel == null)
					$channel = new P10_Channel($args[0]);
				$this->eventHandler->event_chanmessage($user, $channel, $args[1]);
			} else {
				$targetUser = P10_User::getUserByNum($args[0]);
				if($targetUser == null) {
					trigger_error("Server tries to send a privmsg to an user that does not exist or was not found on recv_privmsg.", E_USER_ERROR);
					return;
				}
				$this->eventHandler->event_privmessage($user, $targetUser, $args[1]);
			}
		}
	}
	
	private function recv_notice($from, $args) {
		$user = P10_User::getUserByNum($from);
		if($user == null) {
			trigger_error("Server tries to send a notice from an user that does not exist or was not found on recv_notice.", E_USER_ERROR);
			return;
		}
		if($this->eventHandler) {
			if($args[0][0] == "#") {
				$channel = P10_Channel::getChannelByName($args[0]);
				if($channel == null)
					$channel = new P10_Channel($args[0]);
				$this->eventHandler->event_channotice($user, $channel, $args[1]);
			} else {
				$targetUser = P10_User::getUserByNum($args[0]);
				if($targetUser == null) {
					trigger_error("Server tries to send a notice to an user that does not exist or was not found on recv_notice.", E_USER_ERROR);
					return;
				}
				$this->eventHandler->event_privnotice($user, $targetUser, $args[1]);
			}
		}
	}
	
	private function recv_whois($from, $args) {
		/* [get] ACAAF W AX :NetworkServ */
		$fromUser = P10_User::getUserByNum($from);
		if($fromUser == null) {
			trigger_error("Server tries to send a whois from an user that does not exist or was not found on recv_whois.", E_USER_ERROR);
			return;
		}
		$users=explode(",",$args[1]);
		foreach($users as $nick) {
			$user = P10_User::getUserByNick($nick);
			if(!$user) {
				$this->send("401", $from, $nick);
				continue;
			}
			$nick = $user->getNick();
			$ident = $user->getIdent();
			$hostmask = $user->getHost();
			$modes = $user->getModes();
			if($modes->hasMode('x')) {
				if(($fakehost = $modes->hasMode('f'))) {
					$hostmask = $fakehost;
				} elseif(($account = $modes->hasMode('r'))) {
					$hostmask = $account.".".$this->getSetting("his_usermask");
				}
			}
			$realname = $user->getRealname();
			$this->send("311", $from , $nick, $ident, $hostmask, $realname);
			if(((!$modes->hasMode('n') && !$modes->hasMode('k')) || $from == $user->getNumeric()) && count($user->getChannels()) != 0) {
				$channels = "";
				foreach($user->getChannels() as $channel) {
					$cmodes = $channel->getModes();
					$privs = $channel->getUserPrivs($user);
					if($cmodes->hasMode("s") && !$fromUser->isOnChannel($channel) && $from != $user->getNumeric()) continue;
					if($cmodes->hasMode("u") && ($privs & (P10_Channel::USERPRIV_OPED | P10_Channel::USERPRIV_VOICE)) == 0 && $from != $user->getNumeric()) continue;
					$chanstr = ($channels == "" ? "" : " ");
					$prefix = "";
					if(($privs & P10_Channel::USERPRIV_OPED)) {
						$prefix = "@";
					} else if(($privs & P10_Channel::USERPRIV_VOICE)) {
						$prefix = "+";
					}
					$chanstr .= $prefix.$channel->getName();
					if(strlen($channels) + strlen($chanstr) > 450) {
						$this->send("319", $from, $nick, $channels);
						$channels = $prefix.$channel->getName();
					}
				}
				if($channels != "") {
					$this->send("319", $from, $nick, $channels);
				}
			}
			if($fromUser->getModes()->hasMode("o") || $from == $user->getNumeric() || !$this->getSetting("his_name")) {
				$this->send("312", $from, $nick, $user->getServer()->getName(), $user->getServer()->getDescription());
			} else {
				$this->send("312", $from, $nick, $this->getSetting("his_name"), $this->getSetting("his_desc"));
			}
			if($modes->hasMode("o") && (!$modes->hasMode("H") || $fromUser->getModes()->hasMode("o"))) {
				if($modes->hasMode("S")) {
					if($modes->hasMode("D"))
						$this->send("313", $from, $nick, "is a Network Service");
					else
						$this->send("313", $from, $nick, "is an illegal Network Service - HACKER!");
				} else
					$this->send("313", $from, $nick, "is an IRC Operator");
			}
			if(($auth = $modes->hasMode("r"))) {
				$this->send("330", $from, $nick, $auth);
			}
		}
		$this->send("318", $from, $args[1]);
	}
	
	private function recv_away($from, $args) {
		$user = P10_User::getUserByNum($from);
		if($user == null) {
			trigger_error("Server tries to send an away command from an user that does not exist or was not found on recv_away.", E_USER_ERROR);
			return;
		}
		if(count($args) > 0) {
			$user->setAway($args[0]);
		} else {
			$user->setAway(null);
		}
	}
	
	private function recv_mode($from, $args) {
		$user = P10_User::getUserByNum($from);
		if($user == null) {
			trigger_error("Server tries to send a modechange from an user that does not exist or was not found on recv_mode.", E_USER_ERROR);
			return;
		}
		$modes = implode(" ",array_slice($args,1));
		if($args[0][0] == "#") {
			$channel = P10_Channel::getChannelByName($args[0]);
			if($channel == null)
				$channel = new P10_Channel($args[0]);
			$channel->getModes()->setModes($modes);
			if($this->eventHandler)
				$this->eventHandler->event_chanmode($user, $channel, $modes);
		} else {
			$targetUser = P10_User::getUserByNick($args[0]);
			if($targetUser == null) {
				trigger_error("Server tries to send a mode to an user that does not exist or was not found on recv_mode.", E_USER_ERROR);
				return;
			}
			$targetUser->getModes()->setModes($modes);
			if($this->eventHandler)
				$this->eventHandler->event_usermode($targetUser, $modes);
		}
	}
	
	/********************************************************************************************
	 *                                     SERVER FUNCTIONS                                     *
	 ********************************************************************************************/
	
	private function burst() {
		foreach($this->server->getUsers() as $user) {
			$nick = $user->getNick();
			$connect_time = $user->getConnectTime();
			$ident = $user->getIdent();
			$host = $user->getHost();
			$modes = $user->getModes()->getModeString();
			$ip = Numerics::numericFromIP($user->getIP());
			$numeric = $user->getNumeric();
			$realname = $user->getRealname();
			$this->send("N", $nick, $connect_time, $ident, $host, $modes, $ip, $numeric, $realname);
		}
		foreach(P10_Channel::getChannels() as $channel) {
			$sorted_users = array('ov' => array(), 'o' => array(), 'v' => array(), '-' => array());
			$local_users = false;
			foreach($channel->getUsers() as $user) {
				if(substr($user->getNumeric(), 0, 2) != $this->server->getNumeric()) continue; //skip users that are not on the local server
				$privs = $channel->getUserPrivs($user);
				$strPrivs = "";
				if(($privs & P10_Channel::USERPRIV_OPED)) $strPrivs .= "o";
				if(($privs & P10_Channel::USERPRIV_VOICE)) $strPrivs .= "v";
				if($strPrivs == "") $strPrivs = "-";
				$local_users = true;
				$sorted_users[$strPrivs][] = $user;
			}
			if(!$local_users) continue;
			$userStr = "";
			foreach($sorted_users['-'] as $user) {
				if($userStr != "") $userStr.=",";
				$userStr .= $user->getNumeric();
			}
			foreach($sorted_users['ov'] as $i => $user) {
				if($userStr != "") $userStr.=",";
				$userStr .= $user->getNumeric();
				if($i == 0) $userStr .= ":ov";
			}
			foreach($sorted_users['o'] as $i => $user) {
				if($userStr != "") $userStr.=",";
				$userStr .= $user->getNumeric();
				if($i == 0) $userStr .= ":o";
			}
			foreach($sorted_users['v'] as $i => $user) {
				if($userStr != "") $userStr.=",";
				$userStr .= $user->getNumeric();
				if($i == 0) $userStr .= ":v";
			}
			$banString = "";
			//TODO: Build ban String
			$burstString = "";
			$modeString = $channel->getModes()->getModeString();
			if($modeString != "+") {
				$burstString .= $modeString;
			}
			if($userStr != "") {
				if($burstString != "") $burstString .= " ";
				$burstString .= $userStr;
			}
			if($banString != "") {
				if($burstString != "") $burstString .= " ";
				$burstString .= ":%".$banString;
			}
			$this->send("B", $channel->getName(), $channel->getCreateTime(), $burstString);
		}
		$this->send("EB");
	}
	
	/********************************************************************************************
	 *                                    LOCAL USER FUNCTIONS                                  *
	 ********************************************************************************************/
	
	public function addUser($nick, $ident, $host, $ip, $modes, $realname) {
		$user = P10_User::getUserByNick($nick);
		if($user != null) return ERR_NICK_IN_USE;
		$numeric = substr($this->server->getNumeric(),0,2).Numerics::intToNum($this->last_local_numeric, 3);
		while(P10_User::getUserByNum($numeric)) {
			$this->last_local_numeric++;
			$numeric = substr($this->server->getNumeric(),0,2).Numerics::intToNum($this->last_local_numeric, 3);
		}
		$this->last_local_numeric++;
		$modes = new P10_UserModeSet($modes);
		$ip = new IPAddr($ip);
		$user = new P10_User($nick, $numeric, $this->server, time(), $ident, $host, $ip, $realname, $modes);
		if(($this->flags & self::FLAG_CONNECTED)) {
			$ip = $user->getIP()->getNumeric();
			$this->send("N", $nick, $user->getConnectTime(), $ident, $host, $user->getModes()->getModeString(), $ip, $numeric, $realname);
		}
		return $user;
	}
	
	public function delUser($user, $reason) {
		if(!is_a($user, "P10_User")) return ERR_INVALID_USER;
		if($user->getServer() === $this->server) {
			//local user (QUIT)
			$user->quit($reason);
			if(($this->flags & self::FLAG_CONNECTED)) 
				$this->send("Q", $user->getNumeric(), $reason);
		} else {
			//remote user (KILL)
			if(!($this->flags & self::FLAG_CONNECTED)) 
				return ERR_NOT_CONNECTED;
			if($this->eventHandler)
				$this->eventHandler->event_quit($user, "Killed (".$reason.")");
			$user->quit("Killed (".$reason.")");
			$name = ($this->getSetting('his_name') ? $this->getSetting('his_name') : $this->getSetting('name'));

			$this->send("D", $user->getNumeric(), $name, $reason);
		}
	}
	
	public function join($user, $chanName, $privs = 0) {
		if(!is_a($user, "P10_User") || !($user->getServer() === $this->server))
			return ERR_INVALID_USER;
		if($chanName[0] != "#")
			return ERR_INVALID_CHANNAME;
		$channel = P10_Channel::getChannelByName($chanName);
		if($channel == null)
			$channel = new P10_Channel($chanName);
		$channel->joinUser($user);
		if(($this->flags & self::FLAG_CONNECTED))
			$this->send("J", $user->getNumeric(), $chanName, time(), 0);
		if($privs != 0) {
			$channel->setUserPrivs($user, $privs);
			if(($this->flags & self::FLAG_CONNECTED)) {
				$modestr = "+".(($privs & P10_Channel::USERPRIV_OPED) ? "o" : "").(($privs & P10_Channel::USERPRIV_VOICE) ? "v" : "");
				$modestr .= (($privs & P10_Channel::USERPRIV_OPED) ? " ".$user->getNumeric() : "");
				$modestr .= (($privs & P10_Channel::USERPRIV_VOICE) ? " ".$user->getNumeric() : "");
				$this->send("OM", $user->getNumeric(), $chanName, $modestr);
			}
		}
		if($this->eventHandler)
			$this->eventHandler->event_join($user, $channel, false);
	}
	
	public function part($user, $chanName, $reason) {
		if(!is_a($user, "P10_User") || !($user->getServer() === $this->server))
			return ERR_INVALID_USER;
		if(!((is_string($chanName) && $chanName[0] == "#") || is_a($chanName, "P10_Channel")))
			return ERR_INVALID_CHANNAME;
		if(is_a($chanName, "P10_Channel"))
			$channel = $chanName;
		else
			$channel = P10_Channel::getChannelByName($chanName);
		if($channel == null)
			$channel = new P10_Channel($chanName);
		if(!$user->isOnChannel($channel)) 
			return ERR_NOT_ON_CHANNEL;
		if($this->eventHandler)
			$this->eventHandler->event_part($user, $channel, $reason);
		$channel->partUser($user, $reason);
		if(($this->flags & self::FLAG_CONNECTED))
			$this->send("L", $user->getNumeric(), $chanName, $reason);
	}
	
	public function kick($user, $target, $chanName, $reason) {
		if(!is_a($user, "P10_User") || !($user->getServer() === $this->server))
			return ERR_INVALID_USER;
		if(!is_a($target, "P10_User"))
			return ERR_INVALID_USER;
		if(!((is_string($chanName) && $chanName[0] == "#") || is_a($chanName, "P10_Channel")))
			return ERR_INVALID_CHANNAME;
		if(is_a($chanName, "P10_Channel"))
			$channel = $chanName;
		else
			$channel = P10_Channel::getChannelByName($chanName);
		if($channel == null)
			$channel = new P10_Channel($chanName);
		if(!$target->isOnChannel($channel)) 
			return ERR_NOT_ON_CHANNEL;
		if($this->eventHandler)
			$this->eventHandler->event_kick($user, $target, $channel, $reason);
		$channel->partUser($target, $reason);
		if(($this->flags & self::FLAG_CONNECTED))
			$this->send("K", $user->getNumeric(), $chanName, $target->getNumeric(), $reason);
	}
	
	public function privmsg($user, $target, $message) {
		if(!is_a($user, "P10_User") || !($user->getServer() === $this->server))
			return ERR_INVALID_USER;
		if(!is_a($target, "P10_User") && !is_a($target, "P10_Channel") && !(is_string($target) && ($target[0] == "#" || P10_User::getUserByNick($target))))
			return ERR_INVALID_USER;
		if(is_a($target, "P10_Channel"))
			$targetStr = $target->getName();
		else if(is_a($target, "P10_User"))
			$targetStr = $target->getNumeric();
		else if($target[0] == "#")
			$targetStr = $target;
		else
			$targetStr = P10_User::getUserByNick($target)->getNumeric();
		
		if($this->eventHandler) {
			if($targetStr[0] == "#") {
				$channel = P10_Channel::getChannelByName($targetStr);
				if($channel == null)
					$channel = new P10_Channel($targetStr);
				$this->eventHandler->event_chanmessage($user, $channel, $message);
			} else {
				$targetUser = P10_User::getUserByNum($targetStr);
				$this->eventHandler->event_privmessage($user, $targetUser, $message);
			}
		}
		if(($this->flags & self::FLAG_CONNECTED))
			$this->send("P", $user->getNumeric(), $targetStr, $message);
	}
	
	public function notice($user, $target, $message) {
		if(!is_a($user, "P10_User") || !($user->getServer() === $this->server))
			return ERR_INVALID_USER;
		if(!is_a($target, "P10_User") && !is_a($target, "P10_Channel") && !(is_string($target) && ($target[0] == "#" || P10_User::getUserByNick($target))))
			return ERR_INVALID_USER;
		if(is_a($target, "P10_Channel"))
			$targetStr = $target->getName();
		else if(is_a($target, "P10_User"))
			$targetStr = $target->getNumeric();
		else if($target[0] == "#")
			$targetStr = $target;
		else
			$targetStr = P10_User::getUserByNick($target)->getNumeric();
		
		if($this->eventHandler) {
			if($targetStr[0] == "#") {
				$channel = P10_Channel::getChannelByName($targetStr);
				if($channel == null)
					$channel = new P10_Channel($targetStr);
				$this->eventHandler->event_channotice($user, $channel, $message);
			} else {
				$targetUser = P10_User::getUserByNum($targetStr);
				$this->eventHandler->event_privnotice($user, $targetUser, $message);
			}
		}
		if(($this->flags & self::FLAG_CONNECTED))
			$this->send("O", $user->getNumeric(), $targetStr, $message);
	}
	
	public function mode($user, $target, $modes, $force = false) {
		if(!is_a($user, "P10_User") || !($user->getServer() === $this->server))
			return ERR_INVALID_USER;
		if(!is_a($target, "P10_User") && !is_a($target, "P10_Channel") && !(is_string($target) && ($target[0] == "#" || P10_User::getUserByNick($target))))
			return ERR_INVALID_USER;
		if(is_a($target, "P10_Channel"))
			$targetStr = $target->getName();
		else if(is_a($target, "P10_User"))
			$targetStr = $target->getNumeric();
		else if($target[0] == "#")
			$targetStr = $target;
		else
			$targetStr = P10_User::getUserByNick($target)->getNumeric();
		
		if($targetStr[0] == "#") {
			$channel = P10_Channel::getChannelByName($targetStr);
			if($channel == null)
				$channel = new P10_Channel($targetStr);
			$modes = $channel->getModes()->setModes($modes, true);
			if(($this->flags & self::FLAG_CONNECTED))
				$this->send(($force ? "OM" : "M"), $user->getNumeric(), $targetStr, $modes);
			if($this->eventHandler)
				$this->eventHandler->event_chanmode(($force ? $this->server : $user), $channel, $modes);
		} else {
			$targetUser = P10_User::getUserByNum($targetStr);
			if($targetUser->getServer() === $this->server) {
				//just do it :D
				$modes = $targetUser->getModes()->setModes($modes, true);
				if(($this->flags & self::FLAG_CONNECTED))
					$this->send("M", $targetUser->getNumeric(), $targetUser->getNick(), $modes);
				if($this->eventHandler)
					$this->eventHandler->event_usermode($targetUser, $modes);
			} else {
				//SVSMODE
				if(($this->flags & self::FLAG_CONNECTED))
					$this->send("SM", $user->getNumeric(), $targetUser->getNumeric(), $modes);
			}
		}
	}
	
	
	
}

?>